{ File    : $URL: file:///D:/SVN/svnDelphi/branches/A%20New%20Beginning/VISExt/servicing/src/convert/Crypto.pas $
  Author  : $Author: Deniz Oezmen $
  Changed : $Date: 2021-08-25 23:39:26 +0200 (Mi, 25 Aug 2021) $
  Revision: $Rev: 596 $

  Visionaire Engine extractor cryptographic routines
}
unit Crypto;

interface

uses
  Shared;

procedure Decrypt(const Buffer: TBuffer; Offset, Len: Longint;
  const Key: AnsiString);
function DecryptPNG(const Buffer: TBuffer; PNGName: UTF8String): Boolean;
function DecryptPNGForce(const Buffer: TBuffer): Boolean;
function DecryptWebP(const Buffer: TBuffer; WebPName: UTF8String): Boolean;

implementation

uses
  DCRC,
  MemSearch,
  SdcpMD5,
  SysUtils;

procedure Decrypt(const Buffer: TBuffer; Offset, Len: Longint;
  const Key: AnsiString);
var
  BufIdx,
  KeyIdx: Longint;
begin
  KeyIdx := 1;

  for BufIdx := Offset to Offset + Len - 1 do
  begin
    if (BufIdx < Low(Buffer)) or (BufIdx > High(Buffer)) then
      Break;

    Buffer[BufIdx] := Buffer[BufIdx] xor Ord(Key[KeyIdx]);

    KeyIdx := KeyIdx mod Length(Key) + 1
  end
end;

function DecryptPNG(const Buffer: TBuffer; PNGName: UTF8String): Boolean;
type
  TPNG_IHDR = packed record
    ID             : array[0..3] of Char;
    Width          : Longint;
    Height         : Longint;
    BitDepth       : Byte;
    ColorType      : Byte;
    CompMethod     : Byte;
    FilterMethod   : Byte;
    InterlaceMethod: Byte;
  end;
var
  Pattern  : array[0..3] of Char;
  Idx      : Longint;
  TestBuf  : TBuffer;
  WideName : WideString;
  AnsiName : AnsiString;
  MD5      : TDCP_md5;
  Digest   : array[0..3] of Longword;
  Key      : AnsiString;
  CRCVal,
  TargetCRC: Longword;

  function ValidatePNG: Boolean;
  begin
    // try the decryption ...
    SwitchEndian(Digest[0], True);
    Key := AnsiLowerCase(Format('%.8x', [Digest[0]]));
    TestBuf[4]  := TestBuf[4]  xor Ord(Key[1]);
    TestBuf[5]  := TestBuf[5]  xor Ord(Key[2]);
    TestBuf[6]  := TestBuf[6]  xor Ord(Key[3]);
    TestBuf[7]  := TestBuf[7]  xor Ord(Key[4]);
    TestBuf[8]  := TestBuf[8]  xor Ord(Key[5]);
    TestBuf[9]  := TestBuf[9]  xor Ord(Key[6]);
    TestBuf[10] := TestBuf[10] xor Ord(Key[7]);
    TestBuf[11] := TestBuf[11] xor Ord(Key[8]);

    // ... and validate the chunk checksum
    CRCVal := $ffffffff;
    CalcCRC32(@TestBuf[0], Length(TestBuf), CRCVal);
    CRCVal := not CRCVal;

    // if the key was correct, the test buffer contains the decrypted data
    Result := (CRCVal = TargetCRC)
  end;

begin
  // The width and height values of the PNG files stored in the archives of
  // Visionaire engine games have been encrypted. The encryption method is
  // basically the same as the one used to obfuscate the file table structure,
  // but the encryption key varies from file to file based on the file name.
  Result := False;

  if Length(Buffer) < SizeOf(TPNG_IHDR) + SizeOf(TargetCRC) then
  begin
    WriteLn('    warning: data too short for a PNG file');
    Exit
  end;

  // search the IHDR chunk
  Pattern := 'IHDR';
  Idx := SearchMem(SearchMethod_KnuthMorrisPratt, @Pattern[0], @Buffer[0],
    Length(Pattern), Length(Buffer));

  if (Idx < 0) or (Idx + SizeOf(TPNG_IHDR) + SizeOf(TargetCRC) >
    Length(Buffer)) then
  begin
    WriteLn('    warning: no valid IHDR found');
    Exit
  end;

  // get the targetted CRC value of the IHDR chunk
  Move(Buffer[Idx + SizeOf(TPNG_IHDR)], TargetCRC, SizeOf(TargetCRC));
  SwitchEndian(TargetCRC, True);

  // compute the decryption key from the prefix part of the file name
  PNGName := ExtractFileName(PNGName);
  PNGName := Copy(PNGName, 1, Length(PNGName) -
    Length(ExtractFileExt(PNGName)));

  MD5 := TDCP_md5.Create;

  // first try: use the prefix as a regular unaltered string
  SetLength(TestBuf, SizeOf(TPNG_IHDR));
  Move(Buffer[Idx], TestBuf[0], Length(TestBuf));

  MD5.Init;
  MD5.Update(PNGName[1], Length(PNGName));
  MD5.Final(Digest);

  Result := ValidatePNG;
  if not Result then
  begin
    // if the regular string failed, try the wide variant of the prefix
    SetLength(TestBuf, SizeOf(TPNG_IHDR));
    Move(Buffer[Idx], TestBuf[0], Length(TestBuf));

    WideName := Utf8Decode(PNGName);
    MD5.Init;
    MD5.Update(WideName[1], Length(WideName) * 2);
    MD5.Final(Digest);

    Result := ValidatePNG;
  end;

  if not Result then
  begin
    // if the wide string failed, try the ANSI variant of the prefix
    SetLength(TestBuf, SizeOf(TPNG_IHDR));
    Move(Buffer[Idx], TestBuf[0], Length(TestBuf));

    AnsiName := Utf8ToAnsi(PNGName);
    MD5.Init;
    MD5.Update(AnsiName[1], Length(AnsiName));
    MD5.Final(Digest);

    Result := ValidatePNG;
  end;

  if Result then
    Move(TestBuf[0], Buffer[Idx], Length(TestBuf))
  else
    WriteLn('    warning: no valid PNG encryption key derivable from file nam' +
      'e');

  MD5.Free
end;

function DecryptPNGForce(const Buffer: TBuffer): Boolean;
type
  TPNG_IHDR = packed record
    ID             : array[0..3] of Char;
    Width          : Longint;
    Height         : Longint;
    BitDepth       : Byte;
    ColorType      : Byte;
    CompMethod     : Byte;
    FilterMethod   : Byte;
    InterlaceMethod: Byte;
  end;
var
  Pattern  : array[0..3] of Char;
  Idx      : Longint;
  TestBuf  : TBuffer;
  CurKey1,
  CurKey2  : Word;
  KeyFound : Boolean;
  Hex1,
  Hex2     : string[2];
  BaseCRC,
  InCRC,
  CRCVal,
  TargetCRC: Longword;
begin
  WriteLn('    info: trying to find PNG encryption key by brute force');

  // The width and height values of the PNG files stored in the archives of
  // Visionaire engine games have been encrypted. The encryption method is
  // basically the same as the one used to obfuscate the file table structure,
  // but the encryption key varies from file to file based on the file name.
  // If the correct key could not be derived from the file name, using various
  // assumptions (and the fact that each chunk of a PNG file is protected with
  // a CRC) we can brute-force it in reasonable time:
  // - The probability of an image having a dimension larger than 65535 pixels
  //   is low, so we assume that half of the bytes we want to decrypt are zero
  //   anyway. Thus, there are only 4 encrypted bytes left.
  // - The decryption takes place with a string of hex characters, i.e. there
  //   are only 16 possible partial keys per encrypted byte. Together with the
  //   first assumption, we get a key space of 65536 possibilites. Testing each
  //   of these possibilities with a CRC check afterwards takes only fractions
  //   of a second on contemporary machines.
  Result := False;

  if Length(Buffer) < SizeOf(TPNG_IHDR) + SizeOf(TargetCRC) then
  begin
    WriteLn('    warning: data too short for a PNG file');
    Exit
  end;

  // search the IHDR chunk
  Pattern := 'IHDR';
  Idx := SearchMem(SearchMethod_KnuthMorrisPratt, @Pattern[0], @Buffer[0],
    Length(Pattern), Length(Buffer));

  if (Idx < 0) or (Idx + SizeOf(TPNG_IHDR) + SizeOf(TargetCRC) >
    Length(Buffer)) then
  begin
    WriteLn('    warning: no valid IHDR found');
    Exit
  end;

  // get the targetted CRC value of the IHDR chunk
  Move(Buffer[Idx + SizeOf(TPNG_IHDR)], TargetCRC, SizeOf(TargetCRC));
  SwitchEndian(TargetCRC, True);

  SetLength(TestBuf, SizeOf(TPNG_IHDR));
  Move(Buffer[Idx], TestBuf[0], Length(TestBuf));

  // we assume that no image has a dimension larger than 65535
  TestBuf[4] := $00;
  TestBuf[5] := $00;
  TestBuf[8] := $00;
  TestBuf[9] := $00;

  // precompute the CRC value for as many fixed bytes as possible
  BaseCRC := $ffffffff;
  CalcCRC32(@TestBuf[0], 6, BaseCRC);

  // test all assumed key combinations
  KeyFound := False;
  for CurKey1 := 0 to 255 do
  begin
    // generate hex characters from the first part of the candidate key
    Hex1 := IntToHex(CurKey1, 2);

    // convert to lower case if necessary
    if Ord(Hex1[1]) >= $41 then
      Hex1[1] := Chr(Ord(Hex1[1]) + $20);
    if Ord(Hex1[2]) >= $41 then
      Hex1[2] := Chr(Ord(Hex1[2]) + $20);

    // try the first part of the decryption
    TestBuf[6] := TestBuf[6] xor Ord(Hex1[1]);
    TestBuf[7] := TestBuf[7] xor Ord(Hex1[2]);

    // compute the intermediate CRC
    InCRC := BaseCRC;
    CalcCRC32(@TestBuf[6], 4, InCRC);

    // search for second part of the key
    for CurKey2 := 0 to 255 do
    begin
      // generate hex characters from the second part of the candidate key
      Hex2 := IntToHex(CurKey2, 2);

      if Ord(Hex2[1]) >= $41 then
        Hex2[1] := Chr(Ord(Hex2[1]) + $20);
      if Ord(Hex2[2]) >= $41 then
        Hex2[2] := Chr(Ord(Hex2[2]) + $20);

      // try the second part of the decryption ...
      TestBuf[10] := TestBuf[10] xor Ord(Hex2[1]);
      TestBuf[11] := TestBuf[11] xor Ord(Hex2[2]);

      // ... and validate the chunk checksum
      CRCVal := InCRC;
      CalcCRC32(@TestBuf[10], Length(TestBuf) - 10, CRCVal);
      CRCVal := not CRCVal;

      // check whether the correct key has been found
      if CRCVal = TargetCRC then
      begin
        KeyFound := True;
        Break
      end;

      // get the original data back for the next test
      Move(Buffer[Idx + 10], TestBuf[10], 2)
    end;

    if KeyFound then
      Break;

    // get the original data back for the next test
    Move(Buffer[Idx + 6], TestBuf[6], 2)
  end;

  // if a key has been found, the test buffer contains the decrypted data
  if KeyFound then
  begin
    Move(TestBuf[0], Buffer[Idx], Length(TestBuf));
    Result := True
  end
  else
    WriteLn('    warning: no decryption key found')
end;

function DecryptWebP(const Buffer: TBuffer; WebPName: UTF8String): Boolean;
var
  Pattern : array[0..3] of Char;
  Idx     : Longint;
  TestBuf : TBuffer;
  WideName: WideString;
  AnsiName: AnsiString;
  MD5     : TDCP_md5;
  Digest  : array[0..3] of Longword;
  Key     : AnsiString;
  Extended: Boolean;
  LZ77    : Boolean;

  function ValidateWebP: Boolean;
  begin
    // try the decryption ...
    SwitchEndian(Digest[0], True);
    Key := AnsiLowerCase(Format('%.8x', [Digest[0]]));
    TestBuf[0] := TestBuf[0] xor Ord(Key[1]);
    TestBuf[1] := TestBuf[1] xor Ord(Key[2]);
    TestBuf[2] := TestBuf[2] xor Ord(Key[3]);
    TestBuf[3] := TestBuf[3] xor Ord(Key[4]);
    TestBuf[4] := TestBuf[4] xor Ord(Key[5]);
    TestBuf[5] := TestBuf[5] xor Ord(Key[6]);
    TestBuf[6] := TestBuf[6] xor Ord(Key[7]);
    TestBuf[7] := TestBuf[7] xor Ord(Key[8]);

    // ... and at least try to validate the header sync value for regular WebP
    // files, the signature byte for LZ77-compressed WebP files or the reserved
    // bytes for extended WebP files, respectively
    // (We could also try and validate the VP8 chunk size for regular WebP files
    // or the compressed stream size for LZ77-compressed WebP files, but let's
    // hope this is enough for now.)
    if Extended then
      Result := ((TestBuf[1] = $00) and
                 (TestBuf[2] = $00) and
                 (TestBuf[3] = $00))
    else if LZ77 then
      Result := (TestBuf[4] = $2f)
    else
      Result := (TestBuf[7] = $9d)
  end;

begin
  Result := False;

  // The WebP encryption is a variant of the PNG encryption: It affects eight
  // bytes as well, which are either part of the VP8X chunk, the VP8L chunk or
  // the VP8 chunk.
  // In the latter case, four of these bytes are the chunk size of the VP8
  // chunk. The other four are the start of the image frame, including the first
  // byte of the sync value.
  // Note: I currently don't have a brute force approach for this. It should be
  // possible to construct one for the chunk size part, since this is easily
  // determined, but I haven't delved enough into the WebP format to think of
  // one for the three bytes preceeding the first sync value byte. Nevertheless,
  // the standard approach below seems to work well enough.
  // If the encryption affects the VP8X chunk, it basically scrambles some flag
  // bits and part of the canvas size. However, candidate keys can be reasonably
  // well verified since the encrypted part also contains some reserved bytes
  // which are expected to be zero.
  // If the encrytion affects the VP8L chunk, the size of the compressed stream,
  // a signature byte and part of the stream itself are scrambled. We just check
  // for the signature byte here, though it should be possible to additionally
  // check the stream size.
  if Length(Buffer) < 24 then
  begin
    WriteLn('    warning: data too short for a WebP file');
    Exit
  end;

  Extended := False;
  LZ77 := False;

  // Search for the VP8X chunk first. If this one exists, we're looking at an
  // extended WebP, where the encryption affects the VP8X chunk instead of the
  // VP8 chunk.
  Pattern := 'VP8X';
  Idx := SearchMem(SearchMethod_KnuthMorrisPratt, @Pattern[0], @Buffer[0],
    Length(Pattern), Length(Buffer));

  if (Idx >= 0) and (Idx + 8 <= Length(Buffer)) then
    Extended := True
  else
  begin
    // Search for the VP8L chunk next.
    Pattern := 'VP8L';
    Idx := SearchMem(SearchMethod_KnuthMorrisPratt, @Pattern[0], @Buffer[0],
      Length(Pattern), Length(Buffer));

    if (Idx >= 0) and (Idx + 12 <= Length(Buffer)) then
      LZ77 := True
    else
    begin
      // search the VP8 chunk
      Pattern := 'VP8 ';
      Idx := SearchMem(SearchMethod_KnuthMorrisPratt, @Pattern[0], @Buffer[0],
        Length(Pattern), Length(Buffer));

      if (Idx < 0) or (Idx + 12 > Length(Buffer)) then
      begin
        WriteLn('    warning: no valid VP8 chunk found');
        Exit
      end
    end;
  end;

  Inc(Idx, 4);

  // compute the decryption key from the prefix part of the file name
  WebPName := ExtractFileName(WebPName);
  WebPName := Copy(WebPName, 1, Length(WebPName) -
    Length(ExtractFileExt(WebPName)));

  MD5 := TDCP_md5.Create;

  // first try: use the prefix as a regular unaltered string
  SetLength(TestBuf, 8);
  Move(Buffer[Idx], TestBuf[0], Length(TestBuf));

  MD5.Init;
  MD5.Update(WebPName[1], Length(WebPName));
  MD5.Final(Digest);

  Result := ValidateWebP;
  if not Result then
  begin
    // if the regular string failed, try the wide variant of the prefix
    SetLength(TestBuf, 8);
    Move(Buffer[Idx], TestBuf[0], Length(TestBuf));

    WideName := Utf8Decode(WebPName);
    MD5.Init;
    MD5.Update(WideName[1], Length(WideName) * 2);
    MD5.Final(Digest);

    Result := ValidateWebP;
  end;

  if not Result then
  begin
    // if the wide string failed, try the ANSI variant of the prefix
    SetLength(TestBuf, 8);
    Move(Buffer[Idx], TestBuf[0], Length(TestBuf));

    AnsiName := Utf8ToAnsi(WebPName);
    MD5.Init;
    MD5.Update(AnsiName[1], Length(AnsiName));
    MD5.Final(Digest);

    Result := ValidateWebP;
  end;

  if Result then
    Move(TestBuf[0], Buffer[Idx], Length(TestBuf))
  else
    WriteLn('    warning: no valid WebP encryption key derivable from file na' +
      'me');

  MD5.Free
end;

end.
